import copy
import time

import numpy as np

from os import walk
from scipy.io import loadmat
from pathlib import Path


def put_permittivity_in_ucell(ucell, mat_list, mat_table, wl, type_complex=np.complex128):

    res = np.zeros(ucell.shape, dtype=type_complex)

    for z in range(ucell.shape[0]):
        for y in range(ucell.shape[1]):
            for x in range(ucell.shape[2]):
                material = mat_list[ucell[z, y, x]]
                if type(material) == str:
                    res[z, y, x] = find_nk_index(material, mat_table, wl) ** 2
                else:
                    res[z, y, x] = material ** 2

    return res


def put_permittivity_in_ucell_object(ucell_size, mat_list, obj_list, mat_table, wl,
                                     type_complex=np.complex128):
    # TODO: under development
    res = np.zeros(ucell_size, dtype=type_complex)

    for material, obj_index in zip(mat_list, obj_list):
        if type(material) == str:
            res[obj_index] = find_nk_index(material, mat_table, wl) ** 2
        else:
            res[obj_index] = material ** 2

    return res


def find_nk_index(material, mat_table, wl):
    if material[-6:] == '__real':
        material = material[:-6]
        n_only = True
    else:
        n_only = False

    mat_data = mat_table[material.upper()]

    n_index = np.interp(wl, mat_data[:, 0], mat_data[:, 1])

    if n_only:
        return n_index

    k_index = np.interp(wl, mat_data[:, 0], mat_data[:, 2])
    nk = n_index + 1j * k_index

    return nk


def read_material_table(nk_path=None):
    mat_table = {}

    if nk_path is None:
        nk_path = str(Path(__file__).resolve().parent.parent) + '/nk_data'

    full_path_list, name_list, _ = [], [], []
    for (dirpath, dirnames, filenames) in walk(nk_path):
        full_path_list.extend([f'{dirpath}/{filename}' for filename in filenames])
        name_list.extend(filenames)
    for path, name in zip(full_path_list, name_list):
        if name[-3:] == 'txt':
            data = np.loadtxt(path, skiprows=1)
            mat_table[name[:-4].upper()] = data

        elif name[-3:] == 'mat':
            data = loadmat(path)
            data = np.array([data['WL'], data['n'], data['k']])[:, :, 0].T
            mat_table[name[:-4].upper()] = data
    return mat_table


def cell_compression(cell, type_complex=np.complex128):

    if type_complex == np.complex128:
        type_float = np.float64
    else:
        type_float = np.float32

    # find discontinuities in x
    step_y, step_x = 1. / np.array(cell.shape, dtype=type_float)
    x = []
    y = []
    cell_x = []
    cell_xy = []

    cell_next = np.roll(cell, -1, axis=1)

    for col in range(cell.shape[1]):
        if not (cell[:, col] == cell_next[:, col]).all() or (col == cell.shape[1] - 1):
            x.append(step_x * (col + 1))
            cell_x.append(cell[:, col])

    cell_x = np.array(cell_x).T
    cell_x_next = np.roll(cell_x, -1, axis=0)

    for row in range(cell_x.shape[0]):
        if not (cell_x[row, :] == cell_x_next[row, :]).all() or (row == cell_x.shape[0] - 1):
            y.append(step_y * (row + 1))
            cell_xy.append(cell_x[row, :])

    x = np.array(x).reshape((-1, 1))
    y = np.array(y).reshape((-1, 1))
    cell_comp = np.array(cell_xy)

    return cell_comp, x, y


def fft_piecewise_constant(cell, fourier_order, type_complex=np.complex128):

    if cell.shape[0] == 1:
        fourier_order = [0, fourier_order]
    else:
        fourier_order = [fourier_order, fourier_order]
    cell, x, y = cell_compression(cell, type_complex=type_complex)

    # X axis
    cell_next_x = np.roll(cell, -1, axis=1)
    cell_diff_x = cell_next_x - cell

    modes = np.arange(-2 * fourier_order[1], 2 * fourier_order[1] + 1, 1)

    f_coeffs_x = cell_diff_x @ np.exp(-1j * 2 * np.pi * x @ modes[None, :], dtype=type_complex)
    c = f_coeffs_x.shape[1] // 2

    # x_next = np.vstack(np.roll(x, -1, axis=0)[:-1]) - x
    x_next = np.vstack((np.roll(x, -1, axis=0)[:-1], 1)) - x

    f_coeffs_x[:, c] = (cell @ np.vstack((x[0], x_next[:-1]))).flatten()
    mask = np.ones(f_coeffs_x.shape[1], dtype=bool)
    mask[c] = False
    f_coeffs_x[:, mask] /= (1j * 2 * np.pi * modes[mask])

    # Y axis
    f_coeffs_x_next_y = np.roll(f_coeffs_x, -1, axis=0)
    f_coeffs_x_diff_y = f_coeffs_x_next_y - f_coeffs_x

    modes = np.arange(-2 * fourier_order[0], 2 * fourier_order[0] + 1, 1)

    f_coeffs_xy = f_coeffs_x_diff_y.T @ np.exp(-1j * 2 * np.pi * y @ modes[None, :], dtype=type_complex)
    c = f_coeffs_xy.shape[1] // 2

    y_next = np.vstack((np.roll(y, -1, axis=0)[:-1], 1)) - y

    f_coeffs_xy[:, c] = f_coeffs_x.T @ np.vstack((y[0], y_next[:-1])).flatten()

    if c:
        mask = np.ones(f_coeffs_xy.shape[1], dtype=bool)
        mask[c] = False
        f_coeffs_xy[:, mask] /= (1j * 2 * np.pi * modes[mask])

    return f_coeffs_xy.T


def to_conv_mat(pmt, fourier_order, type_complex=np.complex128):
    # TODO: do conv and 1/conv in simultaneously?
    if len(pmt.shape) == 2:
        print('shape is 2')
        raise ValueError
    ff = 2 * fourier_order + 1

    if pmt.shape[1] == 1:  # 1D

        res = np.zeros((pmt.shape[0], ff, ff)).astype(type_complex)

        for i, layer in enumerate(pmt):
            f_coeffs = fft_piecewise_constant(layer, fourier_order, type_complex=type_complex)

            center = f_coeffs.shape[1] // 2
            conv_idx = np.arange(-ff + 1, ff, 1, dtype=int)
            conv_idx = circulant(conv_idx)
            e_conv = f_coeffs[0, center + conv_idx]
            res[i] = e_conv

    else:  # 2D
        # attention on the order of axis (Z Y X)

        res = np.zeros((pmt.shape[0], ff ** 2, ff ** 2)).astype(type_complex)

        for i, layer in enumerate(pmt):
            pmtvy_fft = fft_piecewise_constant(layer, fourier_order, type_complex=type_complex)

            center = np.array(pmtvy_fft.shape) // 2

            conv_idx = np.arange(-ff + 1, ff, 1)
            # conv_idx = circulant(conv_idx)[ff - 1:, :ff]
            conv_idx = circulant(conv_idx)

            conv_i = np.repeat(conv_idx, ff, axis=1)
            conv_i = np.repeat(conv_i, [ff] * ff, axis=0)
            conv_j = np.tile(conv_idx, (ff, ff))
            res[i] = pmtvy_fft[center[0] + conv_i, center[1] + conv_j]

    # import matplotlib.pyplot as plt
    #
    # plt.figure()
    # plt.imshow(abs(res[0]), cmap='jet')
    # plt.colorbar()
    # plt.show()
    #
    return res


def circulant(c):

    center = c.shape[0] // 2
    circ = np.zeros((center + 1, center + 1), dtype=int)

    for r in range(center+1):
        idx = np.arange(r, r - center - 1, -1, dtype=int)

        assign_value = c[center + idx]
        circ[r] = assign_value

    return circ



from copy import deepcopy

import numpy as np
import torch

from .scattering_method import scattering_1d_1, scattering_1d_2, scattering_1d_3, scattering_2d_1, scattering_2d_wv, \
    scattering_2d_2, scattering_2d_3
from .transfer_method import transfer_1d_1, transfer_1d_2, transfer_1d_3, transfer_1d_conical_1, transfer_1d_conical_2, \
    transfer_1d_conical_3, transfer_2d_1, transfer_2d_wv, transfer_2d_2, transfer_2d_3


class _BaseRCWA:
    def __init__(self, grating_type, n_I=1., n_II=1., theta=0., phi=0., psi=0., fourier_order=10,
                 period=0.7, wavelength=np.linspace(0.5, 2.3, 400), pol=0,
                 patterns=None, ucell=None, ucell_materials=None, thickness=None, algo='TMM', perturbation=1E-10,
                 device='cpu', type_complex=torch.complex128):

        self.device = device
        self.type_complex = type_complex

        # common
        self.grating_type = grating_type  # 1D=0, 1D_conical=1, 2D=2
        self.n_I = n_I
        self.n_II = n_II

        self.theta = torch.tensor(theta * np.pi / 180)
        self.phi = torch.tensor(phi * np.pi / 180)
        self.psi = torch.tensor(psi * np.pi / 180)  # TODO: integrate psi and pol

        self.pol = pol  # TE 0, TM 1
        if self.pol == 0:  # TE
            self.psi = torch.tensor(90 * np.pi / 180, device=self.device)
        elif self.pol == 1:  # TM
            self.psi = torch.tensor(0 * np.pi / 180, device=self.device)
        else:
            print('not implemented yet')
            raise ValueError

        self.fourier_order = fourier_order
        self.ff = 2 * self.fourier_order + 1

        self.period = deepcopy(period)

        self.wavelength = wavelength

        self.patterns = patterns
        self.ucell = deepcopy(ucell)
        self.ucell_materials = ucell_materials
        self.thickness = deepcopy(thickness)

        self.algo = algo
        self.perturbation = perturbation

        self.layer_info_list = []
        self.T1 = None

        self.kx_vector = None

    def get_kx_vector(self):

        k0 = 2 * np.pi / self.wavelength
        fourier_indices = torch.arange(-self.fourier_order, self.fourier_order + 1, device=self.device)
        if self.grating_type == 0:
            kx_vector = k0 * (self.n_I * torch.sin(self.theta) - fourier_indices * (self.wavelength / self.period[0])
                              ).type(self.type_complex)
        else:
            kx_vector = k0 * (self.n_I * torch.sin(self.theta) * torch.cos(self.phi) - fourier_indices * (
                    self.wavelength / self.period[0])).type(self.type_complex)

        idx = torch.nonzero(kx_vector == 0)
        if len(idx):
            # TODO: need imaginary part?
            # TODO: make imaginary part sign consistent
            kx_vector[idx] = self.perturbation
            print('varphi divide by 0: adding perturbation')

        self.kx_vector = kx_vector

    def solve_1d(self, wl, E_conv_all, o_E_conv_all):

        self.layer_info_list = []
        self.T1 = None

        fourier_indices = torch.arange(-self.fourier_order, self.fourier_order + 1, device=self.device)

        delta_i0 = torch.zeros(self.ff, device=self.device, dtype=self.type_complex)
        delta_i0[self.fourier_order] = 1

        k0 = 2 * np.pi / wl

        if self.algo == 'TMM':
            kx_vector, Kx, k_I_z, k_II_z, f, YZ_I, g, inc_term, T \
                = transfer_1d_1(self.ff, self.pol, k0, self.n_I, self.n_II, self.kx_vector,
                                self.theta, delta_i0, self.fourier_order,
                                device=self.device, type_complex=self.type_complex)
        elif self.algo == 'SMM':
            Kx, Wg, Vg, Kzg, Wr, Vr, Kzr, Wt, Vt, Kzt, Ar, Br, Sg \
                = scattering_1d_1(k0, self.n_I, self.n_II, self.theta, self.phi, fourier_indices, self.period,
                                  self.pol, wl=wl)
        else:
            raise ValueError

        count = min(len(E_conv_all), len(o_E_conv_all), len(self.thickness))

        # From the last layer
        for layer_index in range(count)[::-1]:

            E_conv = E_conv_all[layer_index]
            o_E_conv = o_E_conv_all[layer_index]
            d = self.thickness[layer_index]

            if self.pol == 0:
                E_conv_i = None
                A = Kx ** 2 - E_conv
                eigenvalues, W = torch.linalg.eig(A)
                q = eigenvalues ** 0.5

                Q = torch.diag(q)
                V = W @ Q

            elif self.pol == 1:
                E_conv_i = torch.linalg.inv(E_conv)
                B = Kx @ E_conv_i @ Kx - torch.eye(E_conv.shape[0], device=self.device, dtype=self.type_complex)
                o_E_conv_i = torch.linalg.inv(o_E_conv)

                eigenvalues, W = torch.linalg.eig(o_E_conv_i @ B)
                q = eigenvalues ** 0.5

                Q = torch.diag(q)
                V = o_E_conv @ W @ Q

            else:
                raise ValueError

            if self.algo == 'TMM':
                X, f, g, T, a_i, b = transfer_1d_2(k0, q, d, W, V, f, g, self.fourier_order, T,
                                                   device=self.device, type_complex=self.type_complex)

                layer_info = [E_conv_i, q, W, X, a_i, b, d]
                self.layer_info_list.append(layer_info)

            elif self.algo == 'SMM':
                A, B, S_dict, Sg = scattering_1d_2(W, Wg, V, Vg, d, k0, Q, Sg)
            else:
                raise ValueError

        if self.algo == 'TMM':
            de_ri, de_ti, T1 = transfer_1d_3(g, YZ_I, f, delta_i0, inc_term, T, k_I_z, k0, self.n_I, self.n_II,
                                             self.theta, self.pol, k_II_z)
            self.T1 = T1

        elif self.algo == 'SMM':
            de_ri, de_ti = scattering_1d_3(Wt, Wg, Vt, Vg, Sg, self.ff, Wr, self.fourier_order, Kzr, Kzt,
                                           self.n_I, self.n_II, self.theta, self.pol)
        else:
            raise ValueError

        return de_ri, de_ti

    # TODO: scattering method
    def solve_1d_conical(self, wl, E_conv_all, o_E_conv_all):

        self.layer_info_list = []
        self.T1 = None

        fourier_indices = torch.arange(-self.fourier_order, self.fourier_order + 1, device=self.device)

        delta_i0 = torch.zeros(self.ff, device=self.device, dtype=self.type_complex)
        delta_i0[self.fourier_order] = 1

        k0 = 2 * np.pi / wl

        if self.algo == 'TMM':
            Kx, ky, k_I_z, k_II_z, varphi, Y_I, Y_II, Z_I, Z_II, big_F, big_G, big_T \
                = transfer_1d_conical_1(self.ff, k0, self.n_I, self.n_II, self.kx_vector, self.theta, self.phi,
                                        device=self.device, type_complex=self.type_complex)
        elif self.algo == 'SMM':
            print('SMM for 1D conical is not implemented')
            return np.nan, np.nan
        else:
            raise ValueError

        count = min(len(E_conv_all), len(o_E_conv_all), len(self.thickness))

        # From the last layer
        for layer_index in range(count)[::-1]:

            E_conv = E_conv_all[layer_index]
            o_E_conv = o_E_conv_all[layer_index]
            d = self.thickness[layer_index]

        # for e_conv, o_e_conv, d in zip(E_conv_all[::-1], o_E_conv_all[::-1], self.thickness[::-1]):
            E_conv_i = torch.linalg.inv(E_conv)
            o_E_conv_i = torch.linalg.inv(o_E_conv)

            if self.algo == 'TMM':
                big_X, big_F, big_G, big_T, big_A_i, big_B, W_1, W_2, V_11, V_12, V_21, V_22, q_1, q_2\
                    = transfer_1d_conical_2(k0, Kx, ky, E_conv, E_conv_i, o_E_conv_i, self.ff, d,
                                                            varphi, big_F, big_G, big_T,
                                                            device=self.device, type_complex=self.type_complex)

                layer_info = [E_conv_i, q_1, q_2, W_1, W_2, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d]
                self.layer_info_list.append(layer_info)

            elif self.algo == 'SMM':
                raise ValueError
            else:
                raise ValueError

        if self.algo == 'TMM':
            de_ri, de_ti, big_T1 = transfer_1d_conical_3(big_F, big_G, big_T, Z_I, Y_I, self.psi, self.theta, self.ff,
                                                 delta_i0, k_I_z, k0, self.n_I, self.n_II, k_II_z,
                                                 device=self.device, type_complex=self.type_complex)
            self.T1 = big_T1

        elif self.algo == 'SMM':
            raise ValueError
        else:
            raise ValueError

        return de_ri, de_ti

    def solve_2d(self, wavelength, E_conv_all, o_E_conv_all):

        self.layer_info_list = []
        self.T1 = None

        fourier_indices = torch.arange(-self.fourier_order, self.fourier_order + 1, device=self.device)

        delta_i0 = torch.zeros((self.ff ** 2, 1), device=self.device, dtype=self.type_complex)
        delta_i0[self.ff ** 2 // 2, 0] = 1

        I = torch.eye(self.ff ** 2, device=self.device, dtype=self.type_complex)
        O = torch.zeros((self.ff ** 2, self.ff ** 2), device=self.device, dtype=self.type_complex)

        center = self.ff ** 2

        k0 = 2 * np.pi / wavelength

        if self.algo == 'TMM':
            kx_vector, ky_vector, Kx, Ky, k_I_z, k_II_z, varphi, Y_I, Y_II, Z_I, Z_II, big_F, big_G, big_T \
                = transfer_2d_1(self.ff, k0, self.n_I, self.n_II, self.kx_vector, self.period, fourier_indices,
                                self.theta, self.phi, wavelength, device=self.device, type_complex=self.type_complex)
        elif self.algo == 'SMM':
            Kx, Ky, kz_inc, Wg, Vg, Kzg, Wr, Vr, Kzr, Wt, Vt, Kzt, Ar, Br, Sg \
                = scattering_2d_1(self.n_I, self.n_II, self.theta, self.phi, k0, self.period, self.fourier_order)
        else:
            raise ValueError

        count = min(len(E_conv_all), len(o_E_conv_all), len(self.thickness))

        # From the last layer
        for layer_index in range(count)[::-1]:

            E_conv = E_conv_all[layer_index]
            o_E_conv = o_E_conv_all[layer_index]
            d = self.thickness[layer_index]

            E_conv_i = torch.linalg.inv(E_conv)
            o_E_conv_i = torch.linalg.inv(o_E_conv)

            if self.algo == 'TMM':  # TODO: MERGE W V part
                W, V, q = transfer_2d_wv(self.ff, Kx, E_conv_i, Ky, o_E_conv_i, E_conv,
                                         device=self.device, type_complex=self.type_complex)

                big_X, big_F, big_G, big_T, big_A_i, big_B, \
                W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22 \
                    = transfer_2d_2(k0, d, W, V, center, q, varphi, I, O, big_F, big_G, big_T, device=self.device,
                                    type_complex=self.type_complex)

                layer_info = [E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d]
                self.layer_info_list.append(layer_info)

            elif self.algo == 'SMM':
                W, V, LAMBDA = scattering_2d_wv(self.ff, Kx, Ky, E_conv, o_E_conv, o_E_conv_i, E_conv_i)
                A, B, Sl_dict, Sg_matrix, Sg = scattering_2d_2(W, Wg, V, Vg, d, k0, Sg, LAMBDA)
            else:
                raise ValueError

        if self.algo == 'TMM':
            de_ri, de_ti, big_T1 = transfer_2d_3(center, big_F, big_G, big_T, Z_I, Y_I, self.psi, self.theta, self.ff,
                                                 delta_i0, k_I_z, k0, self.n_I, self.n_II, k_II_z, device=self.device,
                                                 type_complex=self.type_complex)
            self.T1 = big_T1

        elif self.algo == 'SMM':
            de_ri, de_ti = scattering_2d_3(Wt, Wg, Vt, Vg, Sg, Wr, Kx, Ky, Kzr, Kzt, kz_inc, self.n_I,
                                           self.pol, self.theta, self.phi, self.fourier_order, self.ff)
        else:
            raise ValueError

        return de_ri.reshape((self.ff, self.ff)).real, de_ti.reshape((self.ff, self.ff)).real



import meent as ee
import matplotlib.pyplot as plt

from scipy.linalg import expm


def field_distribution(grating_type, *args, **kwargs):
    if grating_type == 0:
        res = field_dist_1d(*args, **kwargs)
    else:
        res = field_dist_2d(*args, **kwargs)
    return res


def field_dist_1d(wavelength, n_I, theta, fourier_order, T1, layer_info_list, period, pol, resolution=(100, 1, 100)):

    k0 = 2 * ee.pi / wavelength
    fourier_indices = ee.arange(-fourier_order, fourier_order + 1)

    kx_vector = k0 * (n_I * ee.sin(theta) - fourier_indices * (wavelength / period[0])).astype('complex')
    Kx = ee.diag(kx_vector / k0)

    resolution_z, resolution_y, resolution_x = resolution

    field_cell = ee.zeros((resolution_z * len(layer_info_list), resolution_y, resolution_x, 3), dtype='complex')

    T_layer = T1

    # From the first layer
    for idx_layer, (E_conv_i, q, W, X, a_i, b, d) in enumerate(layer_info_list[::-1]):

        c1 = T_layer[:, None]
        c2 = b @ a_i @ X @ T_layer[:, None]

        Q = ee.diag(q)

        if pol == 0:
            V = W @ Q

        else:
            V = E_conv_i @ W @ Q
            EKx = E_conv_i @ Kx

        for k in range(resolution_z):
            z = k / resolution_z * d

            if pol == 0:  # TE
                Sy = W @ (expm(-k0 * Q * z) @ c1 + expm(k0 * Q * (z - d)) @ c2)
                Ux = V @ (-expm(-k0 * Q * z) @ c1 + expm(k0 * Q * (z - d)) @ c2)
                f_here = (-1j) * Kx @ Sy

                for j in range(resolution_y):
                    for i in range(resolution_x):
                        x = i * period[0] / resolution_x

                        Ey = Sy.T @ ee.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Hx = -1j * Ux.T @ ee.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Hz = f_here.T @ ee.exp(-1j * kx_vector.reshape((-1, 1)) * x)

                        field_cell[resolution_z * idx_layer + k, j, i] = Ey, Hx, Hz
            else:  # TM
                Uy = W @ (expm(-k0 * Q * z) @ c1 + expm(k0 * Q * (z - d)) @ c2)
                Sx = V @ (-expm(-k0 * Q * z) @ c1 + expm(k0 * Q * (z - d)) @ c2)

                f_here = (-1j) * EKx @ Uy

                for j in range(resolution_y):
                    for i in range(resolution_x):
                        x = i * period[0] / resolution_x

                        Hy = Uy.T @ ee.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Ex = 1j * Sx.T @ ee.exp(-1j * kx_vector.reshape((-1, 1)) * x)
                        Ez = f_here.T @ ee.exp(-1j * kx_vector.reshape((-1, 1)) * x)

                        field_cell[resolution_z * idx_layer + k, j, i] = Hy, Ex, Ez

        T_layer = a_i @ X @ T_layer

    return field_cell


def field_dist_2d(wavelength, n_I, theta, phi, fourier_order, T1, layer_info_list, period, resolution=(100, 100, 100)):
    k0 = 2 * ee.pi / wavelength
    fourier_indices = ee.arange(-fourier_order, fourier_order + 1)
    ff = 2 * fourier_order + 1

    kx_vector = k0 * (n_I * ee.sin(theta) * ee.cos(phi) - fourier_indices * (
            wavelength / period[0])).astype('complex')
    ky_vector = k0 * (n_I * ee.sin(theta) * ee.sin(phi) - fourier_indices * (
            wavelength / period[1])).astype('complex')

    Kx = ee.diag(ee.tile(kx_vector, ff).flatten()) / k0
    Ky = ee.diag(ee.tile(ky_vector.reshape((-1, 1)), ff).flatten()) / k0

    resolution_z, resolution_y, resolution_x = resolution
    field_cell = ee.zeros((resolution_z * len(layer_info_list), resolution_y, resolution_x, 6), dtype='complex')

    T_layer = T1

    big_I = ee.eye((len(T1)))

    # From the first layer
    for idx_layer, (E_conv_i, q, W_11, W_12, W_21, W_22, V_11, V_12, V_21, V_22, big_X, big_A_i, big_B, d)\
            in enumerate(layer_info_list[::-1]):

        c = ee.block([[big_I], [big_B @ big_A_i @ big_X]]) @ T_layer

        ff = len(c) // 4

        c1_plus = c[0*ff:1*ff]
        c2_plus = c[1*ff:2*ff]
        c1_minus = c[2*ff:3*ff]
        c2_minus = c[3*ff:4*ff]

        q1 = q[:len(q)//2]
        q2 = q[len(q)//2:]
        big_Q1 = ee.diag(q1)
        big_Q2 = ee.diag(q2)

        for k in range(resolution_z):
            z = k / resolution_z * d

            Sx = W_11 @ (expm(-k0 * big_Q1 * z) @ c1_plus + expm(k0 * big_Q1 * (z-d)) @ c1_minus) \
                 + W_12 @ (expm(-k0 * big_Q2 * z) @ c2_plus + expm(k0 * big_Q2 * (z-d)) @ c2_minus)

            Sy = W_21 @ (expm(-k0 * big_Q1 * z) @ c1_plus + expm(k0 * big_Q1 * (z-d)) @ c1_minus) \
                 + W_22 @ (expm(-k0 * big_Q2 * z) @ c2_plus + expm(k0 * big_Q2 * (z-d)) @ c2_minus)

            Ux = V_11 @ (-expm(-k0 * big_Q1 * z) @ c1_plus + expm(k0 * big_Q1 * (z-d)) @ c1_minus) \
                 + V_12 @ (-expm(-k0 * big_Q2 * z) @ c2_plus + expm(k0 * big_Q2 * (z-d)) @ c2_minus)

            Uy = V_21 @ (-expm(-k0 * big_Q1 * z) @ c1_plus + expm(k0 * big_Q1 * (z-d)) @ c1_minus) \
                 + V_22 @ (-expm(-k0 * big_Q2 * z) @ c2_plus + expm(k0 * big_Q2 * (z-d)) @ c2_minus)

            Sz = -1j * E_conv_i @ (Kx @ Uy - Ky @ Ux)

            Uz = -1j * (Kx @ Sy - Ky @ Sx)

            for j in range(resolution_y):
                y = j * period[1] / resolution_y

                for i in range(resolution_x):
                    x = i * period[0] / resolution_x

                    exp_K = ee.exp(-1j * kx_vector.reshape((1, -1)) * x) * ee.exp(-1j * ky_vector.reshape((-1, 1)) * y)
                    exp_K = exp_K.flatten()

                    Ex = Sx.T @ exp_K
                    Ey = Sy.T @ exp_K
                    Ez = Sz.T @ exp_K

                    Hx = -1j * Ux.T @ exp_K
                    Hy = -1j * Uy.T @ exp_K
                    Hz = -1j * Uz.T @ exp_K

                    field_cell[resolution_z * idx_layer + k, j, i] = [Ex, Ey, Ez, Hx, Hy, Hz]

        T_layer = big_A_i @ big_X @ T_layer

    return field_cell


def field_plot_zx(field_cell, pol=0, plot_indices=(1, 1, 1, 1, 1, 1), y_slice=0, z_slice=-1, zx=True, yx=True):

    if field_cell.shape[-1] == 6:  # 2D grating
        title = ['2D Ex', '2D Ey', '2D Ez', '2D Hx', '2D Hy', '2D Hz', ]
    else:  # 1D grating
        if pol == 0:  # TE
            title = ['1D Ey', '1D Hx', '1D Hz', ]
        else:  # TM
            title = ['1D Hy', '1D Ex', '1D Ez', ]

    if zx:
        for idx in range(len(title)):
            if plot_indices[idx]:
                plt.imshow((abs(field_cell[:, y_slice, :, idx]) ** 2), cmap='jet', aspect='auto')
                # plt.clim(0, 2)  # identical to caxis([-4,4]) in MATLAB
                plt.colorbar()
                plt.title(title[idx])
                plt.show()
    if yx:
        for idx in range(len(title)):
            if plot_indices[idx]:
                plt.imshow((abs(field_cell[z_slice, :, :, idx]) ** 2), cmap='jet', aspect='auto')
                plt.clim(0, 3.5)  # identical to caxis([-4,4]) in MATLAB
                plt.colorbar()
                plt.title(title[idx])
                plt.show()

    # for idx in range(len(title)):
    #     if plot_indices[idx]:
    #         plt.imshow((abs(field_cell[0, :, :, idx]) ** 2), cmap='jet', aspect='auto')
    #         # plt.clim(0, 1.3)  # identical to caxis([-4,4]) in MATLAB
    #         plt.colorbar()
    #         plt.title(title[idx])
    #         plt.show()
    # for idx in range(len(title)):
    #     if plot_indices[idx]:
    #         plt.imshow((abs(field_cell[-1, :, :, idx]) ** 2), cmap='jet', aspect='auto')
    #         # plt.clim(0, 3.2)  # identical to caxis([-4,4]) in MATLAB
    #         plt.colorbar()
    #         plt.title(title[idx])
    #         plt.show()



import meent.integ.backend.meentpy as ee

from os import walk
from scipy.io import loadmat
from pathlib import Path


def put_permittivity_in_ucell(ucell, mat_list, mat_table, wl):

    res = ee.zeros(ucell.shape, dtype='complex')

    for z in range(ucell.shape[0]):
        for y in range(ucell.shape[1]):
            for x in range(ucell.shape[2]):
                material = mat_list[ucell[z, y, x]]
                if type(material) == str:
                    # res[z, y, x] = find_nk_index(material, mat_table, wavelength) ** 2
                    assign_index = [z, y, x]
                    assign_value = find_nk_index(material, mat_table, wl) ** 2
                    res = ee.assign(res, assign_index, assign_value)

                else:
                    # res[z, y, x] = material ** 2
                    assign_index = [z, y, x]
                    assign_value = material ** 2

                    res = ee.assign(res, assign_index, assign_value)

    return res


def put_permittivity_in_ucell_object(ucell_size, mat_list, obj_list, mat_table, wl):
    # TODO: under development
    res = ee.zeros(ucell_size, dtype='complex')

    for material, obj_index in zip(mat_list, obj_list):
        if type(material) == str:
            res[obj_index] = find_nk_index(material, mat_table, wl) ** 2
        else:
            res[obj_index] = material ** 2

    return res


def find_nk_index(material, mat_table, wl):
    if material[-6:] == '__real':
        material = material[:-6]
        n_only = True
    else:
        n_only = False

    mat_data = mat_table[material.upper()]

    n_index = ee.interp(wl, mat_data[:, 0], mat_data[:, 1])

    if n_only:
        return n_index

    k_index = ee.interp(wl, mat_data[:, 0], mat_data[:, 2])
    nk = n_index + 1j * k_index

    return nk


def read_material_table(nk_path=None):
    mat_table = {}

    if nk_path is None:
        nk_path = str(Path(__file__).resolve().parent.parent) + '/nk_data'

    full_path_list, name_list, _ = [], [], []
    for (dirpath, dirnames, filenames) in walk(nk_path):
        full_path_list.extend([f'{dirpath}/{filename}' for filename in filenames])
        name_list.extend(filenames)
    for path, name in zip(full_path_list, name_list):
        if name[-3:] == 'txt':
            data = ee.loadtxt(path, skiprows=1)
            mat_table[name[:-4].upper()] = data

        elif name[-3:] == 'mat':
            data = loadmat(path)
            data = ee.array([data['WL'], data['n'], data['k']])[:, :, 0].T
            mat_table[name[:-4].upper()] = data
    return mat_table


def cell_compression(cell):
    # find discontinuities in x
    step_y, step_x = 1. / ee.array(cell.shape)
    x = []
    y = []
    cell_x = []
    cell_xy = []

    cell_next = ee.roll(cell, -1, axis=1)

    for col in range(cell.shape[1]):
        if not (cell[:, col] == cell_next[:, col]).all() or (col == cell.shape[1] - 1):
            x.append(step_x * (col + 1))
            cell_x.append(cell[:, col])

    cell_x = ee.array(cell_x).T
    cell_x_next = ee.roll(cell_x, -1, axis=0)

    for row in range(cell_x.shape[0]):
        if not (cell_x[row, :] == cell_x_next[row, :]).all() or (row == cell_x.shape[0] - 1):
            y.append(step_y * (row + 1))
            cell_xy.append(cell_x[row, :])

    x = ee.array(x).reshape((-1, 1))
    y = ee.array(y).reshape((-1, 1))
    cell_comp = ee.array(cell_xy)

    return cell_comp, x, y


def fft_piecewise_constant(cell, fourier_order):
    if cell.shape[0] == 1:
        fourier_order = [0, fourier_order]
    else:
        fourier_order = [fourier_order, fourier_order]
    cell, x, y = cell_compression(cell)

    # X axis
    cell_next_x = ee.roll(cell, -1, axis=1)
    cell_diff_x = cell_next_x - cell

    modes = ee.arange(-2 * fourier_order[1], 2 * fourier_order[1] + 1, 1)

    f_coeffs_x = cell_diff_x @ ee.exp(-1j * 2 * ee.pi * x @ modes[None, :])
    c = f_coeffs_x.shape[1] // 2

    x_next = ee.vstack((ee.roll(x, -1, axis=0)[:-1], 1)) - x

    # f_coeffs_x[:, c] = (cell @ ee.vstack((x[0], x_next[:-1]))).flatten()

    assign_index = [ee.arange(len(f_coeffs_x)), ee.array([c])]
    assign_value = (cell @ ee.vstack((x[0], x_next[:-1]))).flatten()

    f_coeffs_x = ee.assign(f_coeffs_x, assign_index, assign_value)

    mask = ee.ones(f_coeffs_x.shape[1], dtype=bool)
    # mask[c] = False

    mask = ee.assign(mask, c, False)


    # f_coeffs_x[:, mask] /= (1j * 2 * ee.pi * modes[mask])

    assign_index = mask
    assign_value = f_coeffs_x[:, mask] / (1j * 2 * ee.pi * modes[mask])

    f_coeffs_x = ee.assign(f_coeffs_x, assign_index, assign_value, row_all=True)

    # Y axis
    f_coeffs_x_next_y = ee.roll(f_coeffs_x, -1, axis=0)
    f_coeffs_x_diff_y = f_coeffs_x_next_y - f_coeffs_x

    modes = ee.arange(-2 * fourier_order[0], 2 * fourier_order[0] + 1, 1)

    f_coeffs_xy = f_coeffs_x_diff_y.T @ ee.exp(-1j * 2 * ee.pi * y @ modes[None, :])
    c = f_coeffs_xy.shape[1] // 2

    y_next = ee.vstack((ee.roll(y, -1, axis=0)[:-1], 1)) - y

    # f_coeffs_xy[:, c] = f_coeffs_x.T @ ee.vstack((y[0], y_next[:-1])).flatten()

    assign_value = f_coeffs_x.T @ ee.vstack((y[0], y_next[:-1])).flatten()
    f_coeffs_xy = ee.assign(f_coeffs_xy, c, assign_value, row_all=True)

    if c:
        mask = ee.ones(f_coeffs_xy.shape[1], dtype=bool)
        # mask[c] = False
        mask = ee.assign(mask, c, False)

        # f_coeffs_xy[:, mask] /= (1j * 2 * ee.pi * modes[mask])

        assign_value = f_coeffs_xy[:, mask] / (1j * 2 * ee.pi * modes[mask])
        f_coeffs_xy = ee.assign(f_coeffs_xy, mask, assign_value, row_all=True)

    return f_coeffs_xy.T


def to_conv_mat(pmt, fourier_order):

    if len(pmt.shape) == 2:
        print('shape is 2')
        raise ValueError
    ff = 2 * fourier_order + 1

    if pmt.shape[1] == 1:  # 1D

        res = ee.zeros((pmt.shape[0], ff, ff)).astype('complex')

        for i, layer in enumerate(pmt):
            # f_coeffs = fft_piecewise_constant(layer, fourier_order)
            # A = ee.roll(circulant(f_coeffs.flatten()), (f_coeffs.size + 1) // 2, 0)
            # res[i] = A[:2 * fourier_order + 1, :2 * fourier_order + 1]
            f_coeffs = fft_piecewise_constant(layer, fourier_order)

            center = f_coeffs.shape[1] // 2

            conv_idx = ee.arange(-ff + 1, ff, 1)
            conv_idx = circulant(conv_idx)

            e_conv = f_coeffs[0, center + conv_idx]
            # res = res.at[i].set(e_conv)
            res = ee.assign(res, i, e_conv)

    else:  # 2D
        # attention on the order of axis (Z Y X)

        # TODO: separate fourier order
        res = ee.zeros((pmt.shape[0], ff ** 2, ff ** 2)).astype('complex')

        for i, layer in enumerate(pmt):
            # pmtvy_fft = fft_piecewise_constant(layer, fourier_order)
            #
            # center = ee.array(pmtvy_fft.shape) // 2
            #
            # conv_idx = ee.arange(-ff + 1, ff, 1)
            # conv_idx = circulant(conv_idx)[ff - 1:, :ff]
            #
            # conv_i = ee.repeat(conv_idx, ff, axis=1)
            # conv_i = ee.repeat(conv_i, [ff] * ff, axis=0)
            # conv_j = ee.tile(conv_idx, (ff, ff))
            # res[i] = pmtvy_fft[center[0] + conv_i, center[1] + conv_j]
            f_coeffs = fft_piecewise_constant(layer, fourier_order)

            center = ee.array(f_coeffs.shape) // 2

            conv_idx = ee.arange(-ff + 1, ff, 1)

            conv_idx = circulant(conv_idx)

            conv_i = ee.repeat(conv_idx, ff, axis=1)
            conv_i = ee.repeat(conv_i, ff, axis=0)
            conv_j = ee.tile(conv_idx, (ff, ff))

            # res = res.at[i].set(f_coeffs[center[0] + conv_i, center[1] + conv_j])
            assign_value = f_coeffs[center[0] + conv_i, center[1] + conv_j]
            res = ee.assign(res, i, assign_value)
    # import matplotlib.pyplot as plt
    #
    # plt.figure()
    # plt.imshow(abs(res[0]), cmap='jet')
    # plt.colorbar()
    # plt.show()
    #
    return res


# def draw_fill_factor(patterns_fill_factor, grating_type, resolution=1000, mode=0):
#
#     # res in Z X Y
#     if grating_type == 2:
#         res = ee.zeros((len(patterns_fill_factor), resolution, resolution), dtype='complex')
#     else:
#         res = ee.zeros((len(patterns_fill_factor), 1, resolution), dtype='complex')
#
#     if grating_type in (0, 1):  # TODO: handle this by len(fill_factor)
#         # fill_factor is not exactly implemented.
#         for i, (n_ridge, n_groove, fill_factor) in enumerate(patterns_fill_factor):
#             permittivity = ee.ones((1, resolution), dtype='complex')
#             cut = int(resolution * fill_factor)
#             permittivity[0, :cut] *= n_ridge ** 2
#             permittivity[0, cut:] *= n_groove ** 2
#             res[i, 0] = permittivity
#     else:  # 2D
#         for i, (n_ridge, n_groove, fill_factor) in enumerate(patterns_fill_factor):
#             fill_factor = ee.array(fill_factor)
#             permittivity = ee.ones((resolution, resolution), dtype='complex')
#             cut = (resolution * fill_factor)  # TODO: need parenthesis?
#             permittivity *= n_groove ** 2
#             permittivity[:int(cut[1]), :int(cut[0])] *= n_ridge ** 2
#             res[i] = permittivity
#
#     return res

def circulant(c):

    center = ee.array(c.shape) // 2
    circ = ee.zeros((center[0] + 1, center[0] + 1), dtype='int32')

    for r in range(center[0]+1):
        idx = ee.arange(r, r - center - 1, -1)

        # circ = circ.at[r].set(c[center + idx])
        assign_value = c[center + idx]
        circ = ee.assign(circ, r, assign_value)

    return circ



import time
from functools import partial

import jax
import jax.numpy as jnp
import numpy as np

from ._base import _BaseRCWA
from .convolution_matrix import to_conv_mat, put_permittivity_in_ucell, read_material_table, \
    to_conv_mat_piecewise_continuous
from .field_distribution import field_dist_1d, field_dist_1d_conical, field_dist_2d, field_plot


class RCWAJax(_BaseRCWA):
    def __init__(self, n_I=1., n_II=1., theta=0, phi=0, psi=0,
                 period=(100,),
                 wavelength=900, ucell=None,
                 thickness=None, perturbation=1E-10,
                 mode=1, grating_type=0,
                 pol=0, fourier_order=40,
                 ucell_materials=None,
                 algo='TMM',
                 device='cpu', type_complex=jnp.complex128):

        super().__init__(grating_type, n_I, n_II, theta, phi, psi, pol, fourier_order, period, wavelength,
                         ucell, ucell_materials,
                         thickness, algo, perturbation, device, type_complex)

        self.device = device
        self.mode = mode
        self.type_complex = type_complex

        self.mat_table = read_material_table(type_complex=self.type_complex)
        self.layer_info_list = []

    def _tree_flatten(self):
        children = (self.n_I, self.n_II, self.theta, self.phi, self.psi,
                    self.period, self.wavelength, self.ucell, self.thickness, self.perturbation)
        aux_data = {
            'mode': self.mode,
            'grating_type': self.grating_type,
            'pol': self.pol,
            'fourier_order': self.fourier_order,
            'ucell_materials': self.ucell_materials,
            'algo': self.algo,
            'device': self.device,
            'type_complex': self.type_complex,

        }

        return children, aux_data

    @classmethod
    def _tree_unflatten(cls, aux_data, children):
        return cls(*children, **aux_data)

    @jax.jit
    def solve(self, wavelength, e_conv_all, o_e_conv_all):

        self.get_kx_vector()
        # self.kx_vector = self.get_kx_vector()

        if self.grating_type == 0:
            de_ri, de_ti, layer_info_list, T1 = self.solve_1d(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 1:
            de_ri, de_ti, layer_info_list, T1 = self.solve_1d_conical(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 2:
            de_ri, de_ti, layer_info_list, T1 = self.solve_2d(wavelength, e_conv_all, o_e_conv_all)
        else:
            raise ValueError

        self.layer_info_list = layer_info_list
        self.T1 = T1

        return de_ri.real, de_ti.real

    @jax.jit
    def aaa(self, ucell):
        E_conv_all = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        de_ri, de_ti = self.solve(self.wavelength, E_conv_all, o_E_conv_all)
        return de_ri, de_ti

    def run_ucell(self):

        ucell = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)

        E_conv_all = to_conv_mat_piecewise_continuous(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all = to_conv_mat_piecewise_continuous(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        de_ri, de_ti = self.solve(self.wavelength, E_conv_all, o_E_conv_all)

        # de_ri, de_ti = self.aaa(ucell)

        return de_ri, de_ti

    def run_ucell_vmap(self):

        ucell = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        self.wavelength = 901
        ucell1 = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all1 = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all1 = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        self.wavelength = 902
        ucell2 = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all2 = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all2 = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        self.wavelength = 903
        ucell3 = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all3 = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all3 = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        a = jnp.array([900, 901, 902, 903])
        b = jnp.array([E_conv_all, E_conv_all1, E_conv_all2, E_conv_all3])
        c = jnp.array([o_E_conv_all, o_E_conv_all1, o_E_conv_all2, o_E_conv_all3])

        de_ri, de_ti = jax.vmap(self.solve)(a, b, c)

        return de_ri, de_ti

    def run_ucell_pmap(self):

        ucell = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        self.wavelength = 901
        ucell1 = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all1 = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all1 = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        self.wavelength = 902
        ucell2 = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all2 = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all2 = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        self.wavelength = 903
        ucell3 = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength,
                                          type_complex=self.type_complex)
        E_conv_all3 = to_conv_mat(ucell, self.fourier_order, type_complex=self.type_complex)
        o_E_conv_all3 = to_conv_mat(1 / ucell, self.fourier_order, type_complex=self.type_complex)

        a = jnp.array([900, 901, 902, 903])
        b = jnp.array([E_conv_all, E_conv_all1, E_conv_all2, E_conv_all3])
        c = jnp.array([o_E_conv_all, o_E_conv_all1, o_E_conv_all2, o_E_conv_all3])

        de_ri, de_ti = jax.pmap(self.solve)(a, b, c)

        return de_ri, de_ti

    def calculate_field(self, resolution=None, plot=True):

        if self.grating_type == 0:
            resolution = [100, 1, 100] if not resolution else resolution
            field_cell = field_dist_1d(self.wavelength, self.kx_vector, self.n_I, self.theta, self.fourier_order, self.T1,
                                       self.layer_info_list, self.period, self.pol, resolution=resolution,
                                       type_complex=self.type_complex)
        elif self.grating_type == 1:
            resolution = [100, 1, 100] if not resolution else resolution
            field_cell = field_dist_1d_conical(self.wavelength, self.kx_vector, self.n_I, self.theta, self.phi, self.fourier_order, self.T1,
                                               self.layer_info_list, self.period, resolution=resolution,
                                               type_complex=self.type_complex)

        else:
            resolution = [10, 10, 10] if not resolution else resolution
            field_cell = field_dist_2d(self.wavelength, self.kx_vector, self.n_I, self.theta, self.phi, self.fourier_order, self.T1,
                                       self.layer_info_list, self.period, resolution=resolution,
                                       type_complex=self.type_complex)
        if plot:
            field_plot(field_cell, self.pol)
        return field_cell


if __name__ == '__main__':
    pass



import meent.integ.backend

if meent.integ.backend.mode == 2:
    from meent.integ.backend.be_numpy import *
    print(33)
    # import numpy as np
    #
    # backend = 'numpy'
    # linspace = np.linspace
    # pi = np.pi
    # arange = np.arange
    # zeros = np.zeros
    # diag = np.diag
    #
    # linalg = np.linalg
    # eye = np.eye
    # nan = np.nan
    # interp = np.interp
    # loadtxt = np.loadtxt
    # array = np.array
    # roll = np.roll
    # exp = np.exp
    # vstack = np.vstack
    # ones = np.ones
    # repeat = np.repeat
    # tile = np.tile
    #
    # sin = np.sin
    # cos = np.cos
    # block = np.block
    #
    # real = np.real
    # imag = np.imag
    # conj = np.conj
    #
    # nonzero = np.nonzero
    #
    # arctan = np.arctan
    #
    # hstack = np.hstack
    #
    # eig = np.linalg.eig




elif meent.integ.backend.mode == 3:
    from meent.integ.backend.be_jax import *
    print(23)



import numpy as np
import jax
from functools import partial

from jax import tree_util


# @partial(jax.jit, static_argnums=(0, 1))
def call_solver(mode=0, *args, **kwargs):
    """
    decide backend and return RCWA solver instance

    Args:
        mode: decide backend. 0 is numpy and 1 is JAX.
        *args: passed to RCWA instance
        **kwargs: passed to RCWA instance

    Returns:
        RCWA: RCWA solver instance

    """
    if mode == 0:
        from .on_numpy.rcwa import RCWANumpy
        RCWA = RCWANumpy(mode, *args, **kwargs)
    elif mode == 1:
        from .on_jax.rcwa import RCWAJax
        RCWA = RCWAJax(mode=mode, *args, **kwargs)

    elif mode == 2:
        from .on_torch.rcwa import RCWATorch
        RCWA = RCWATorch(mode, *args, **kwargs)
    else:
        raise ValueError

    return RCWA


# def sweep_wavelength(wavelength_array, mode=0, *args, **kwargs):
#     # wavelength = np.linspace(500, 1000, 10)
#     # spectrum_r = []
#     # spectrum_t = []
#     spectrum_r = np.zeros(wavelength_array.shape[0])
#     spectrum_t = np.zeros(wavelength_array.shape[0])
#     solver = call_solver(mode, *args, **kwargs)
#     spectrum_r, spectrum_t = init_spectrum_array(solver.grating_type, wavelength_array, solver.fourier_order)
#
#     for i, wavelength in enumerate(wavelength_array):
#
#         solver.wavelength = np.array([wavelength])
#         de_ri, de_ti = solver.run_ucell()
#         # spectrum_r.append(de_ri)
#         # spectrum_t.append(de_ti)
#         spectrum_r[i] = de_ri
#         spectrum_t[i] = de_ti
#
#     # for i, wavelength in enumerate(wavelength):
#     #     wavelength = np.array([wavelength])
#     #     solver = call_solver(wavelength=wavelength, *args, **kwargs)
#     #     de_ri, de_ti = solver.run_ucell()
#     #     spectrum_r.append(de_ri)
#     #     spectrum_t.append(de_ti)
#     #     # spectrum_r[i] = de_ri
#     #     # spectrum_t[i] = de_ti
#
#     # spectrum_r = np.array(spectrum_r)
#     # spectrum_t = np.array(spectrum_t)
#
#     return spectrum_r, spectrum_t
#
#
# def init_spectrum_array(grating_type, wavelength_array, fourier_order):
#     if grating_type in (0, 1):
#         spectrum_r = np.zeros((len(wavelength_array), 2 * fourier_order + 1))
#         spectrum_t = np.zeros((len(wavelength_array), 2 * fourier_order + 1))
#     elif grating_type == 2:
#         spectrum_r = np.zeros((len(wavelength_array), 2 * fourier_order + 1, 2 * fourier_order + 1))
#         spectrum_t = np.zeros((len(wavelength_array), 2 * fourier_order + 1, 2 * fourier_order + 1))
#     else:
#         raise ValueError
#     return spectrum_r, spectrum_t




import numpy

import jax
import jax.numpy as np
from jax import jit

backend = 'jax'


linspace = np.linspace
pi = np.pi
arange = np.arange
zeros = np.zeros
diag = np.diag

linalg = np.linalg
eye = np.eye
nan = np.nan
interp = np.interp
loadtxt = numpy.loadtxt
array = np.array
roll = np.roll
exp = np.exp
vstack = np.vstack
ones = np.ones
repeat = np.repeat
tile = np.tile

sin = np.sin
cos = np.cos
block = np.block

real = np.real
imag = np.imag
conj = np.conj

nonzero = np.nonzero

arctan = np.arctan

hstack = np.hstack
eig = np.linalg.eig

from jax import jit


# @jit
def _assign_row_all(arr, index, value):
    row, col = arr.shape
    return arr.at[:, index].set(value)


@jit
def _assign_col_all(arr, index, value):
    row, col = arr.shape
    return arr.at[index, np.arange(col)].set(value)


@jit
def _assign(arr, index, value):
    return arr.at[index].set(value)


def assign(arr, index, value, row_all=False, col_all=False):
    if type(index) == list:
        index = tuple(index)

    if row_all:
        arr = _assign_row_all(arr, index, value)
    elif col_all:
        arr = _assign_col_all(arr, index, value)
    else:
        arr = _assign(arr, index, value)
    return arr


def assign1(arr, index, value, row_all=False, col_all=False):
    if type(index) == list:
        index = tuple(index)

    if row_all:
        arr = arr.at[:, index].set(value)
    elif col_all:
        arr = arr.at[index, :].set(value)
    else:
        arr = arr.at[index].set(value)
    return arr



# import numpy
#
# import jax
# import jax.numpy as np
# from jax import jit
#
# backend = 'jax'

#
# linspace = jit(np.linspace)
# pi = np.pi
# arange = jit(np.arange)
# zeros = np.zeros
# diag = jit(np.diag)
#
# linalg = np.linalg
# eye = jit(np.eye)
# nan = np.nan
# interp = jit(np.interp)
# loadtxt = numpy.loadtxt
# array = jit(np.array)
# roll = jit(np.roll)
# exp = jit(np.exp)
# vstack = jit(np.vstack)
# ones = jit(np.ones)
# repeat = jit(np.repeat)
# tile = jit(np.tile)
#
# sin = jit(np.sin)
# cos = jit(np.cos)
# block = jit(np.block)
#
# real = jit(np.real)
# imag = jit(np.imag)
# conj = jit(np.conj)
#
# nonzero = jit(np.nonzero)
#
# arctan = jit(np.arctan)
#
# hstack = jit(np.hstack)
# eig = np.linalg.eig
#
#
# def assign(arr, index, value, row_all=False, col_all=False):
#     if type(index) == list:
#         index = tuple(index)
#
#     if row_all:
#         arr = arr.at[:, index].set(value)
#     elif col_all:
#         arr = arr.at[index, :].set(value)
#     else:
#         arr = arr.at[index].set(value)
#     return arr





import time

from ._base import _BaseRCWA
from .convolution_matrix import to_conv_mat, put_permittivity_in_ucell, read_material_table
from .field_distribution import field_dist_1d, field_dist_2d, field_plot_zx


class RCWAInteg(_BaseRCWA):

    def __init__(self, mode=0, grating_type=0, n_I=1., n_II=1., theta=0, phi=0, psi=0, fourier_order=40, period=(100,),
                 wavelength=900, pol=0, patterns=None, ucell=None, ucell_materials=None, thickness=None, algo='TMM',
                 *args, **kwargs):

        super().__init__(grating_type, n_I, n_II, theta, phi, psi, fourier_order, period, wavelength, pol, patterns, ucell, ucell_materials,
                         thickness, algo)

        self.mode = mode
        self.spectrum_r, self.spectrum_t = None, None
        # self.init_spectrum_array()
        self.mat_table = read_material_table()

    def solve(self, wavelength, e_conv_all, o_e_conv_all):

        # TODO: !handle uniform layer

        if self.grating_type == 0:
            de_ri, de_ti = self.solve_1d(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 1:
            de_ri, de_ti = self.solve_1d_conical(wavelength, e_conv_all, o_e_conv_all)
        elif self.grating_type == 2:
            de_ri, de_ti = self.solve_2d(wavelength, e_conv_all, o_e_conv_all)
        else:
            raise ValueError

        return de_ri.real, de_ti.real

    def run_ucell(self):

        ucell = put_permittivity_in_ucell(self.ucell, self.ucell_materials, self.mat_table, self.wavelength)

        e_conv_all = to_conv_mat(ucell, self.fourier_order)
        o_e_conv_all = to_conv_mat(1 / ucell, self.fourier_order)

        de_ri, de_ti = self.solve(self.wavelength, e_conv_all, o_e_conv_all)

        return de_ri, de_ti

    def calculate_field(self, resolution=None, plot=True):

        if self.grating_type == 0:
            resolution = [100, 1, 100] if not resolution else resolution
            field_cell = field_dist_1d(self.wavelength, self.n_I, self.theta, self.fourier_order, self.T1,
                                self.layer_info_list, self.period, self.pol, resolution=resolution)
        else:
            resolution = [100, 100, 100] if not resolution else resolution
            field_cell = field_dist_2d(self.wavelength, self.n_I, self.theta, self.phi, self.fourier_order, self.T1,
                                self.layer_info_list, self.period, resolution=resolution)

        if plot:
            field_plot_zx(field_cell, self.pol)

        return field_cell




import numpy as np


backend = 'numpy'
linspace = np.linspace
pi = np.pi
arange = np.arange
zeros = np.zeros
diag = np.diag

linalg = np.linalg
eye = np.eye
nan = np.nan
interp = np.interp
loadtxt = np.loadtxt
array = np.array
roll = np.roll
exp = np.exp
vstack = np.vstack
ones = np.ones
repeat = np.repeat
tile = np.tile

sin = np.sin
cos = np.cos
block = np.block

real = np.real
imag = np.imag
conj = np.conj

nonzero = np.nonzero

arctan = np.arctan

hstack = np.hstack

eig = np.linalg.eig


def assign(arr, index, value, row_all=False, col_all=False):
    if type(index) == list:
        index = tuple(index)

    if row_all:
        arr[:, index] = value
    elif col_all:
        arr[index, :] = value
    else:
        arr[index] = value
    return arr



import tensorflow as tf
import numpy as np


backend = 'numpy'
linspace = tf.linspace
pi = tf.constant(np.pi)

arange = tf.experimental.numpy.arange

# zeros = tf.zeros
zeros = tf.experimental.numpy.zeros

# diag = tf.linalg.diag
diag = tf.experimental.numpy.diag


linalg = tf.linalg
eye = tf.eye

# nan = tf.nan
nan = tf.constant(np.nan)

interp = None

loadtxt = np.loadtxt

array = tf.experimental.numpy.array

# roll = tf.roll
roll = tf.experimental.numpy.roll


exp = tf.exp

# vstack = tf.vstack
vstack = tf.experimental.numpy.vstack


ones = tf.ones
repeat = tf.repeat
tile = tf.tile

sin = tf.sin
cos = tf.cos
block = None

real = tf.math.real
imag = tf.math.imag
conj = tf.math.conj

nonzero = tf.experimental.numpy.nonzero

# arctan = tf.math.atan
arctan = tf.experimental.numpy.arctan

hstack = tf.experimental.numpy.hstack

eig = tf.linalg.eig


def assign(arr, index, value, row_all=False, col_all=False):
    if type(index) == list:
        index = tuple(index)

    if row_all:
        arr[:, index] = value
    elif col_all:
        arr[index, :] = value
    else:
        arr[index] = value
    return arr

# https://stackoverflow.com/questions/38420288/how-to-implement-element-wise-1d-interpolation-in-tensorflow
def interp():
    pass


from setuptools import setup, find_packages

setup(
    name='meent',
    version='0.7.0',
    url='https://github.com/kc-ml2/meent',
    author='KC ML2',
    author_email='yongha@kc-ml2.com',
    packages=['meent'] + find_packages(include=['meent.*']),
    install_requires=[
        'numpy==1.23.3',
        'jax==0.4.1',
        'matplotlib==3.5.3',
    ],
    python_requires='>=3.8',
    long_description_content_type="text/markdown",
    package_data={
        'meent': ['nk_data/filmetrics/*.txt', 'nk_data/matlab/*.mat'],
    },
)



import numpy as np

import jax
import jax.numpy as jnp
from functools import partial

from jax.experimental import host_callback
from jax import pure_callback

loadtxt = np.loadtxt

backend = 'jax'

pi = jnp.pi

diag = jax.jit(jnp.diag)

inv = jax.jit(jnp.linalg.inv)

nan = jnp.nan

# TODO
# interp = jax.jit(jnp.interp)
interp = jnp.interp

exp = jax.jit(jnp.exp)
vstack = jax.jit(jnp.vstack)

sin = jax.jit(jnp.sin)
cos = jax.jit(jnp.cos)
block = jax.jit(jnp.block)

real = jax.jit(jnp.real)
imag = jax.jit(jnp.imag)
conj = jax.jit(jnp.conj)

arctan = jax.jit(jnp.arctan)

hstack = jax.jit(jnp.hstack)

array = partial(jax.jit, static_argnums=(1,))(jnp.array)

roll = partial(jax.jit, static_argnums=(2,))(jnp.roll)
arange = partial(jax.jit, static_argnums=(0, 1, 2))(jnp.arange)

ones = partial(jax.jit, static_argnums=(0, 1))(jnp.ones)
zeros = partial(jax.jit, static_argnums=(0, 1))(jnp.zeros)

repeat = partial(jax.jit, static_argnums=(1, 2,))(jnp.repeat)
tile = partial(jax.jit, static_argnums=(1,))(jnp.tile)

linspace = partial(jax.jit, static_argnums=(0, 1, 2))(jnp.linspace)

eye = partial(jax.jit, static_argnums=(0,))(jnp.eye)
nonzero = partial(jax.jit, static_argnums=(0,))(jnp.nonzero)


@partial(jax.jit, static_argnums=(3, 4))
def assign(arr, index, value, row_all=False, col_all=False):
    if type(index) == list:
        index = tuple(index)

    if row_all:
        arr = arr.at[:, index].set(value)
    elif col_all:
        arr = arr.at[index, :].set(value)
    else:
        arr = arr.at[index].set(value)
    return arr


def eig2(matrix, type_complex=jnp.complex128):
    """Wraps jnp.linalg.eig so that it can be jit-ed on a machine with GPUs."""
    eigenvalues_shape = jax.ShapeDtypeStruct(matrix.shape[:-1], type_complex)
    eigenvectors_shape = jax.ShapeDtypeStruct(matrix.shape, type_complex)

    return host_callback.call(
        # We force this computation to be performed on the cpu by jit-ing and
        # explicitly specifying the device.
        jax.jit(jnp.linalg.eig, device=jax.devices('cpu')[0]),
        matrix.astype(type_complex),
        result_shape=(eigenvalues_shape, eigenvectors_shape),
    )


def eig(X, type_complex):
    import numpy as np
    # jax.config.update('jax_enable_x64', True)

    # type_complex = jnp.complex128
    # X = np.array(X, dtype=np.complex64)
    _eig = lambda x: jax.jit(jnp.linalg.eig, device=jax.devices('cpu')[0])(x)

    eigenvalues_shape = jax.ShapeDtypeStruct(X.shape[:-1], type_complex)
    eigenvectors_shape = jax.ShapeDtypeStruct(X.shape, type_complex)

    result_shape_dtype = (eigenvalues_shape, eigenvectors_shape)

    return jax.pure_callback(_eig, result_shape_dtype, X)


# eig = jnp.linalg.eig

# @partial(jax.jit, backend='cpu')
# def eig2(mat):
#     print('eig')
#     return jnp.linalg.eig(mat)

# @partial(jax.jit, static_argnums=(1, ), backend='cpu')
# def eig(mat, _):
#     # _ in args is used for not to jit compile. Refer https://github.com/google/jax/issues/13959#issue-1528545365
#     return jnp.linalg.eig(mat)


# def eig(mat, _):
#     with jax.default_device(jax.devices('cpu')[0]):
#         a=jnp.linalg.eig(mat)
#         return a

fft = jnp.fft



from jax import tree_util

from .rcwa import RCWAJax

tree_util.register_pytree_node(RCWAJax,
                               RCWAJax._tree_flatten,
                               RCWAJax._tree_unflatten)
